//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/utils/utils_gtest.h>

#include "core/problem/problem.h"
#include "gnss/sensor/sensor_gnss.h"
#include "gnss/capture/capture_gnss.h"
#include "gnss/factor/factor_gnss_pseudo_range.h"

using namespace Eigen;
using namespace wolf;
using namespace GnssUtils;

// groundtruth transformations
Vector3d t_ecef_enu, t_enu_map, t_map_base, t_base_antena; // Antena extrinsics
Vector3d t_ecef_antena;
Vector3d rpy_enu_map;
Matrix3d R_ecef_enu, R_enu_map;
Quaterniond q_map_base;
Vector1d clock_drift;

GnssUtils::Satellite sat1({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat2({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat3({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});
GnssUtils::Satellite sat4({SYS_GPS, 1,Eigen::Vector3d::Zero(),Eigen::Vector3d::Zero(),0,0,0});

GnssUtils::Range range1, range2, range3, range4;


// WOLF
ProblemPtr prb = Problem::create("PO", 3);
SolverCeresPtr solver = std::make_shared<SolverCeres>(prb);
SensorGnssPtr gnss_sensor = std::static_pointer_cast<SensorGnss>(prb->installSensor("SensorGnss", "gnss", t_base_antena, std::make_shared<ParamsSensorGnss>()));
FrameBasePtr frm;
CaptureGnssPtr cap;
FeatureGnssSatellitePtr ftr1, ftr2, ftr3, ftr4;
FactorGnssPseudoRangePtr fac1, fac2, fac3, fac4;

void randomGroundtruth()
{
    // ECEF-ENU
    t_ecef_enu = Vector3d::Random() * 1e3;
    Vector3d rpy_ecef_enu = M_PI*Vector3d::Random();
    R_ecef_enu = (AngleAxis<double>(rpy_ecef_enu(0), Vector3d::UnitX()) *
                  AngleAxis<double>(rpy_ecef_enu(1), Vector3d::UnitY()) *
                  AngleAxis<double>(rpy_ecef_enu(2), Vector3d::UnitZ())).toRotationMatrix();

    // ENU-MAP
    t_enu_map = Vector3d::Random() * 1e3;
    rpy_enu_map = M_PI*Vector3d::Random();
    R_enu_map  = (AngleAxis<double>(rpy_enu_map(0), Vector3d::UnitX()) *
                  AngleAxis<double>(rpy_enu_map(1), Vector3d::UnitY()) *
                  AngleAxis<double>(rpy_enu_map(2), Vector3d::UnitZ())).toRotationMatrix();

    // MAP-BASE
    t_map_base = Vector3d::Random() * 1e2;
    q_map_base = Quaterniond::UnitRandom();

    // BASE-ANTENA (EXTRINSICS)
    t_base_antena = Vector3d::Random();

    // composition
    t_ecef_antena = R_ecef_enu * (R_enu_map * (q_map_base * t_base_antena + t_map_base) + t_enu_map ) + t_ecef_enu;

    // SATELLITES
    sat1.pos = Vector3d::Random() * 1e4;
    sat2.pos = Vector3d::Random() * 1e4;
    sat3.pos = Vector3d::Random() * 1e4;
    sat4.pos = Vector3d::Random() * 1e4;

    // clock drift
    clock_drift = Vector1d::Random() * 1e2;

    // pseudo ranges
    range1.P_corrected = (sat1.pos-t_ecef_antena).norm() + clock_drift(0);
    range2.P_corrected = (sat2.pos-t_ecef_antena).norm() + clock_drift(0);
    range3.P_corrected = (sat3.pos-t_ecef_antena).norm() + clock_drift(0);
    range4.P_corrected = (sat4.pos-t_ecef_antena).norm() + clock_drift(0);
    range1.P_var = 1.0;
    range2.P_var = 1.0;
    range3.P_var = 1.0;
    range4.P_var = 1.0;
}

void setUpProblem()
{
    // remove previous setup
    if (frm)
        frm->remove();

    solver->getSolverOptions().max_num_iterations = 100;

    // Sensor
    // ENU-MAP
    gnss_sensor->setEnuMapTranslationState(t_enu_map);
    gnss_sensor->setEnuMapRollState(rpy_enu_map(0));
    gnss_sensor->setEnuMapPitchState(rpy_enu_map(1));
    gnss_sensor->setEnuMapYawState(rpy_enu_map(2));
    // ECEF-ENU
    gnss_sensor->setEnuEcef(R_ecef_enu.transpose(), R_ecef_enu.transpose()*(-t_ecef_enu));
    // Extrinsics
    gnss_sensor->getP()->setState(t_base_antena);

    // Frame
    Vector7d frm_state;
    frm_state.head<3>() = t_map_base;
    frm_state.tail<4>() = q_map_base.coeffs();
    frm = prb->emplaceFrame( TimeStamp(0), frm_state);

    // capture
    cap = CaptureBase::emplace<CaptureGnss>(frm, TimeStamp(0), gnss_sensor, nullptr);
    cap->getStateBlock('T')->unfix();
    cap->getStateBlock('T')->setState(clock_drift);

    // features
    ftr1 = FeatureBase::emplace<FeatureGnssSatellite>(cap, obsd_t(), sat1, range1); // obsd_t data is not used in pseudo range factors
    ftr2 = FeatureBase::emplace<FeatureGnssSatellite>(cap, obsd_t(), sat2, range2); // obsd_t data is not used in pseudo range factors
    ftr3 = FeatureBase::emplace<FeatureGnssSatellite>(cap, obsd_t(), sat3, range3); // obsd_t data is not used in pseudo range factors
    ftr4 = FeatureBase::emplace<FeatureGnssSatellite>(cap, obsd_t(), sat4, range4); // obsd_t data is not used in pseudo range factors

    // factors
    fac1 = FactorBase::emplace<FactorGnssPseudoRange>(ftr1, ftr1, gnss_sensor, nullptr, false);
    fac2 = FactorBase::emplace<FactorGnssPseudoRange>(ftr2, ftr2, gnss_sensor, nullptr, false);
    fac3 = FactorBase::emplace<FactorGnssPseudoRange>(ftr3, ftr3, gnss_sensor, nullptr, false);
    fac4 = FactorBase::emplace<FactorGnssPseudoRange>(ftr4, ftr4, gnss_sensor, nullptr, false);

    // ASSERTS
    // ENU-MAP
    ASSERT_FLOAT_EQ(gnss_sensor->getEnuMapRoll()->getState()(0), rpy_enu_map(0));
    ASSERT_FLOAT_EQ(gnss_sensor->getEnuMapPitch()->getState()(0), rpy_enu_map(1));
    ASSERT_FLOAT_EQ(gnss_sensor->getEnuMapYaw()->getState()(0), rpy_enu_map(2));
    ASSERT_MATRIX_APPROX(gnss_sensor->getEnuMapTranslation()->getState(), t_enu_map, 1e-3);
    // extrinsics
    ASSERT_MATRIX_APPROX(gnss_sensor->getP()->getState(), t_base_antena, 1e-3);
    // Frame
    ASSERT_MATRIX_APPROX(frm->getP()->getState(), t_map_base, 1e-3);
    ASSERT_MATRIX_APPROX(frm->getO()->getState(), q_map_base.coeffs(), 1e-3);
    // clock drift
    ASSERT_MATRIX_APPROX(clock_drift, cap->getStateBlock('T')->getState(), 1e-3);
    // composition
    ASSERT_MATRIX_APPROX(gnss_sensor->computeFrameAntennaPosEcef(frm), t_ecef_antena, 1e-3);
}

void fixAllStates()
{
    // ENU-MAP
    gnss_sensor->getEnuMapRoll()->fix();
    gnss_sensor->getEnuMapPitch()->fix();
    gnss_sensor->getEnuMapYaw()->fix();
    gnss_sensor->getEnuMapTranslation()->fix();
    // Antena
    gnss_sensor->getP()->fix();
    // Frame
    frm->fix();
    // clock drift
    cap->fix();
 }

////////////////////////////////////////////////////////
TEST(FactorGnssPreusoRangeTest, observe_clock_drift)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        cap->getStateBlock('T')->unfix();

        // perturb
        cap->getStateBlock('T')->perturb(1e2);

        // Only 1 factor
        fac2->setStatus(FAC_INACTIVE);
        fac3->setStatus(FAC_INACTIVE);
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(clock_drift, cap->getStateBlock('T')->getState(), 1e-3);
    }
}

TEST(FactorGnssPreusoRangeTest, observe_frame_p)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        frm->getP()->unfix();

        // perturb
        frm->getP()->perturb(1);

        // Only 3 factors
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(frm->getP()->getState(), t_map_base, 1e-3);
    }
}

TEST(FactorGnssPreusoRangeTest, observe_frame_p_clock)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        frm->getP()->unfix();
        cap->getStateBlock('T')->unfix();

        // perturb
        frm->getP()->perturb(1);
        cap->getStateBlock('T')->perturb(1e2);

        // all 4 factors

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(frm->getP()->getState(), t_map_base, 1e-3);
        ASSERT_MATRIX_APPROX(clock_drift, cap->getStateBlock('T')->getState(), 1e-3);
    }
}

TEST(FactorGnssPreusoRangeTest, observe_enumap_p)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        gnss_sensor->getEnuMapTranslation()->unfix();

        // perturb
        gnss_sensor->getEnuMapTranslation()->perturb(1e2);

        // Only 3 factors
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(gnss_sensor->getEnuMapTranslation()->getState(), t_enu_map, 1e-3);
    }
}

TEST(FactorGnssPreusoRangeTest, observe_enumap_o)
{
    for (auto i = 0; i < 100; i++)
    {
        // setup random problem
        randomGroundtruth();
        setUpProblem();

        // fix/unfix
        fixAllStates();
        gnss_sensor->getEnuMapRoll()->unfix();
        gnss_sensor->getEnuMapPitch()->unfix();
        gnss_sensor->getEnuMapYaw()->unfix();

        // perturb
        gnss_sensor->getEnuMapRoll()->perturb(1);
        gnss_sensor->getEnuMapPitch()->perturb(1);
        gnss_sensor->getEnuMapYaw()->perturb(1);

        // Only 3 factors
        fac4->setStatus(FAC_INACTIVE);

        // solve
        std::string report = solver->solve(SolverManager::ReportVerbosity::FULL);
        //std::cout << report << std::endl;

        ASSERT_MATRIX_APPROX(gnss_sensor->getEnuMapTranslation()->getState(), t_enu_map, 1e-3);
    }
}

int main(int argc, char **argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
